@@ -17,19 +17,19 @@
 /* This implements 'user defined functions' */
 
 /*
-** Known bugs:
-**
-** Memory for functions are never freed!
-** Shared libraries are not closed before mysqld exists;
-**   - This is because we can't be sure if some threads is using
-**     a functions.
-**
-** The buggs only affects applications that creates and frees a lot of
-** dynamic functions, so this shouldn't be a real problem.
+   Known bugs:
+  
+   Memory for functions is never freed!
+   Shared libraries are not closed before mysqld exits;
+     - This is because we can't be sure if some threads are using
+       a function.
+  
+   The bugs only affect applications that create and free a lot of
+   dynamic functions, so this shouldn't be a real problem.
 */
 
-#ifdef __GNUC__
-#pragma implementation				// gcc: implement sql_udf.h
+#ifdef USE_PRAGMA_IMPLEMENTATION
+#pragma implementation				// gcc: Class implementation
 #endif
 
 #include "mysql_priv.h"
@@ -110,15 +110,15 @@ static char *init_syms(udf_func *tmp, char *nm)
   */
   if (!tmp->func_init && !tmp->func_deinit && tmp->type != UDFTYPE_AGGREGATE)
   {
-    if (opt_allow_suspicious_udfs)
-      sql_print_error(ER(ER_CANT_FIND_DL_ENTRY), nm);
-    else
+    if (!opt_allow_suspicious_udfs)
       return nm;
+    if (current_thd->variables.log_warnings)
+      sql_print_warning(ER(ER_CANT_FIND_DL_ENTRY), nm);
   }
-
   return 0;
 }
 
+
 extern "C" byte* get_hash_key(const byte *buff,uint *length,
 			      my_bool not_used __attribute__((unused)))
 {
@@ -127,9 +127,10 @@ extern "C" byte* get_hash_key(const byte *buff,uint *length,
   return (byte*) udf->name.str;
 }
 
+
 /*
-** Read all predeclared functions from mysql.func and accept all that
-** can be used.
+  Read all predeclared functions from mysql.func and accept all that
+  can be used.
 */
 
 void udf_init()
@@ -158,12 +159,13 @@ void udf_init()
     DBUG_VOID_RETURN;
   }
   initialized = 1;
+  new_thd->thread_stack= (char*) &new_thd;
   new_thd->store_globals();
   new_thd->db= my_strdup("mysql", MYF(0));
   new_thd->db_length=5;
 
   bzero((gptr) &tables,sizeof(tables));
-  tables.alias= tables.real_name= (char*) "func";
+  tables.alias= tables.table_name= (char*) "func";
   tables.lock_type = TL_READ;
   tables.db=new_thd->db;
 
@@ -185,7 +187,7 @@ void udf_init()
     char *dl_name= get_field(&mem, table->field[2]);
     bool new_dl=0;
     Item_udftype udftype=UDFTYPE_FUNCTION;
-    if (table->fields >= 4)			// New func table
+    if (table->s->fields >= 4)			// New func table
       udftype=(Item_udftype) table->field[3]->val_int();
 
     /*
@@ -193,7 +195,9 @@ void udf_init()
       This is done to ensure that only approved dll from the system
       directories are used (to make this even remotely secure).
     */
-    if (strchr(dl_name, '/') || name.length > NAME_LEN)
+    if (strchr(dl_name, '/') ||
+        IF_WIN(strchr(dl_name, '\\'),0) ||
+        strlen(name.str) > NAME_LEN)
     {
       sql_print_error("Invalid row in mysql.func table for function '%.64s'",
                       name.str);
@@ -222,7 +226,7 @@ void udf_init()
     }
     tmp->dlhandle = dl;
     {
-      char buf[MAX_FIELD_NAME+16], *missing;
+      char buf[NAME_LEN+16], *missing;
       if ((missing= init_syms(tmp, buf)))
       {
         sql_print_error(ER(ER_CANT_FIND_DL_ENTRY), missing);
@@ -304,6 +308,10 @@ static void del_udf(udf_func *udf)
 void free_udf(udf_func *udf)
 {
   DBUG_ENTER("free_udf");
+  
+  if (!initialized)
+    DBUG_VOID_RETURN;
+
   rw_wrlock(&THR_LOCK_udf);
   if (!--udf->usage_count)
   {
@@ -328,6 +336,9 @@ udf_func *find_udf(const char *name,uint length,bool mark_used)
   udf_func *udf=0;
   DBUG_ENTER("find_udf");
 
+  if (!initialized)
+    DBUG_RETURN(NULL);
+
   /* TODO: This should be changed to reader locks someday! */
   if (mark_used)
     rw_wrlock(&THR_LOCK_udf);  /* Called during fix_fields */
@@ -400,7 +411,7 @@ int mysql_create_function(THD *thd,udf_func *udf)
 
   if (!initialized)
   {
-    send_error(thd, ER_OUT_OF_RESOURCES, ER(ER_OUT_OF_RESOURCES));
+    my_message(ER_OUT_OF_RESOURCES, ER(ER_OUT_OF_RESOURCES), MYF(0));
     DBUG_RETURN(1);
   }
 
@@ -409,50 +420,49 @@ int mysql_create_function(THD *thd,udf_func *udf)
     This is done to ensure that only approved dll from the system
     directories are used (to make this even remotely secure).
   */
-  if (strchr(udf->dl, '/'))
+  if (strchr(udf->dl, '/') || IF_WIN(strchr(udf->dl, '\\'),0))
   {
-    send_error(thd, ER_UDF_NO_PATHS,ER(ER_UDF_NO_PATHS));
+    my_message(ER_UDF_NO_PATHS, ER(ER_UDF_NO_PATHS), MYF(0));
     DBUG_RETURN(1);
   }
   if (udf->name.length > NAME_LEN)
   {
-    net_printf(thd, ER_TOO_LONG_IDENT,udf->name);
+    my_error(ER_TOO_LONG_IDENT, MYF(0), udf->name);
     DBUG_RETURN(1);
   }
 
   rw_wrlock(&THR_LOCK_udf);
   if ((hash_search(&udf_hash,(byte*) udf->name.str, udf->name.length)))
   {
-    net_printf(thd, ER_UDF_EXISTS, udf->name);
+    my_error(ER_UDF_EXISTS, MYF(0), udf->name);
     goto err;
   }
   if (!(dl = find_udf_dl(udf->dl)))
   {
+    DBUG_PRINT("info", ("Calling dlopen, udf->dl: %s", udf->dl));
     if (!(dl = dlopen(udf->dl, RTLD_NOW)))
     {
       DBUG_PRINT("error",("dlopen of %s failed, error: %d (%s)",
 			  udf->dl,errno,dlerror()));
-      net_printf(thd, ER_CANT_OPEN_LIBRARY, udf->dl, errno, dlerror());
+      my_error(ER_CANT_OPEN_LIBRARY, MYF(0),
+                      udf->dl, errno, dlerror());
       goto err;
     }
     new_dl=1;
   }
   udf->dlhandle=dl;
   {
-    char buf[MAX_FIELD_NAME+16], *missing;
+    char buf[NAME_LEN+16], *missing;
     if ((missing= init_syms(udf, buf)))
     {
-      net_printf(thd, ER_CANT_FIND_DL_ENTRY, missing);
+      my_error(ER_CANT_FIND_DL_ENTRY, MYF(0), missing);
       goto err;
     }
   }
   udf->name.str=strdup_root(&mem,udf->name.str);
   udf->dl=strdup_root(&mem,udf->dl);
   if (!(u_d=add_udf(&udf->name,udf->returns,udf->dl,udf->type)))
-  {
-    send_error(thd,0);		// End of memory
     goto err;
-  }
   u_d->dlhandle = dl;
   u_d->func=udf->func;
   u_d->func_init=udf->func_init;
@@ -464,23 +474,23 @@ int mysql_create_function(THD *thd,udf_func *udf)
 
   bzero((char*) &tables,sizeof(tables));
   tables.db= (char*) "mysql";
-  tables.real_name= tables.alias= (char*) "func";
+  tables.table_name= tables.alias= (char*) "func";
   /* Allow creation of functions even if we can't open func table */
   if (!(table = open_ltable(thd,&tables,TL_WRITE)))
     goto err;
 
-  restore_record(table,default_values);		// Default values for fields
+  restore_record(table, s->default_values);	// Default values for fields
   table->field[0]->store(u_d->name.str, u_d->name.length, system_charset_info);
-  table->field[1]->store((longlong) u_d->returns);
+  table->field[1]->store((longlong) u_d->returns, TRUE);
   table->field[2]->store(u_d->dl,(uint) strlen(u_d->dl), system_charset_info);
-  if (table->fields >= 4)			// If not old func format
-    table->field[3]->store((longlong) u_d->type);
+  if (table->s->fields >= 4)			// If not old func format
+    table->field[3]->store((longlong) u_d->type, TRUE);
   error = table->file->write_row(table->record[0]);
 
   close_thread_tables(thd);
   if (error)
   {
-    net_printf(thd, ER_ERROR_ON_WRITE, "mysql.func",error);
+    my_error(ER_ERROR_ON_WRITE, MYF(0), "mysql.func", error);
     del_udf(u_d);
     goto err;
   }
@@ -503,14 +513,14 @@ int mysql_drop_function(THD *thd,const LEX_STRING *udf_name)
   DBUG_ENTER("mysql_drop_function");
   if (!initialized)
   {
-    send_error(thd, ER_OUT_OF_RESOURCES, ER(ER_OUT_OF_RESOURCES));
+    my_message(ER_OUT_OF_RESOURCES, ER(ER_OUT_OF_RESOURCES), MYF(0));
     DBUG_RETURN(1);
   }
   rw_wrlock(&THR_LOCK_udf);  
   if (!(udf=(udf_func*) hash_search(&udf_hash,(byte*) udf_name->str,
 				    (uint) udf_name->length)))
   {
-    net_printf(thd, ER_FUNCTION_NOT_DEFINED, udf_name->str);
+    my_error(ER_FUNCTION_NOT_DEFINED, MYF(0), udf_name->str);
     goto err;
   }
   del_udf(udf);
@@ -523,7 +533,7 @@ int mysql_drop_function(THD *thd,const LEX_STRING *udf_name)
 
   bzero((char*) &tables,sizeof(tables));
   tables.db=(char*) "mysql";
-  tables.real_name= tables.alias= (char*) "func";
+  tables.table_name= tables.alias= (char*) "func";
   if (!(table = open_ltable(thd,&tables,TL_WRITE)))
     goto err;
   table->field[0]->store(udf_name->str, udf_name->length, system_charset_info);
