@@ -360,9 +360,20 @@ sub mtr_kill_leftovers () {
 
   # First, kill all masters and slaves that would conflict with
   # this run. Make sure to remove the PID file, if any.
+  # FIXME kill IM manager first, else it will restart the servers, how?!
 
   my @args;
 
+  for ( my $idx; $idx < 2; $idx++ )
+  {
+    push(@args,{
+                pid      => 0,          # We don't know the PID
+                pidfile  => $::instance_manager->{'instances'}->[$idx]->{'path_pid'},
+                sockfile => $::instance_manager->{'instances'}->[$idx]->{'path_sock'},
+                port     => $::instance_manager->{'instances'}->[$idx]->{'port'},
+               });
+  }
+
   for ( my $idx; $idx < 2; $idx++ )
   {
     push(@args,{
@@ -454,8 +465,7 @@ sub mtr_kill_leftovers () {
 
         if ( kill(0, @pids) )           # Check if some left
         {
-          # FIXME maybe just mtr_warning() ?
-          mtr_error("can't kill process(es) " . join(" ", @pids));
+          mtr_warning("can't kill process(es) " . join(" ", @pids));
         }
       }
     }
@@ -468,7 +478,7 @@ sub mtr_kill_leftovers () {
   {
     if ( mtr_ping_mysqld_server($srv->{'port'}, $srv->{'sockfile'}) )
     {
-      mtr_error("can't kill old mysqld holding port $srv->{'port'}");
+      mtr_warning("can't kill old mysqld holding port $srv->{'port'}");
     }
   }
 }
@@ -770,7 +780,15 @@ sub mtr_record_dead_children () {
 }
 
 sub start_reap_all {
-  $SIG{CHLD}= 'IGNORE';                 # FIXME is this enough?
+  # This causes terminating processes to not become zombies, avoiding
+  # the need for (or possibility of) explicit waitpid().
+  $SIG{CHLD}= 'IGNORE';
+
+  # On some platforms (Linux, QNX, OSX, ...) there is potential race
+  # here. If a process terminated before setting $SIG{CHLD} (but after
+  # any attempt to waitpid() it), it will still be a zombie. So we
+  # have to handle any such process here.
+  while(waitpid(-1, &WNOHANG) > 0) { };
 }
 
 sub stop_reap_all {
@@ -824,7 +842,7 @@ sub sleep_until_file_created ($$$) {
     }
 
     # Check if it died after the fork() was successful 
-    if ( waitpid($pid,&WNOHANG) == $pid )
+    if ( $pid > 0 && waitpid($pid,&WNOHANG) == $pid )
     {
       return 0;
     }
@@ -848,14 +866,16 @@ sub sleep_until_file_created ($$$) {
 sub mtr_kill_processes ($) {
   my $pids = shift;
 
-  foreach my $sig (15,9)
+  foreach my $sig (15, 9)
   {
-    my $retries= 20;                    # FIXME 20 seconds, this is silly!
-    kill($sig, @{$pids});
-    while ( $retries-- and  kill(0, @{$pids}) )
+    my $retries= 10;
+    while (1)
     {
-      mtr_debug("Sleep 1 second waiting for processes to die");
-      sleep(1)                      # Wait one second
+      kill($sig, @{$pids});
+      last unless kill (0, @{$pids}) and $retries--;
+
+      mtr_debug("Sleep 2 second waiting for processes to die");
+      sleep(2);
     }
   }
 }
@@ -877,7 +897,14 @@ sub mtr_exit ($) {
 #  cluck("Called mtr_exit()");
   mtr_timer_stop_all($::glob_timers);
   local $SIG{HUP} = 'IGNORE';
-  kill('HUP', -$$);
+  # ToDo: Signalling -$$ will only work if we are the process group
+  # leader (in fact on QNX it will signal our session group leader,
+  # which might be Do-compile or Pushbuild, causing tests to be
+  # aborted). So we only do it if we are the group leader. We might
+  # set ourselves as the group leader at startup (with
+  # POSIX::setpgrp(0,0)), but then care must be needed to always do
+  # proper child process cleanup.
+  kill('HUP', -$$) if $$ == getpgrp();
   sleep 2;
   exit($code);
 }
