@@ -64,16 +64,21 @@ ClusterMgr::ClusterMgr(TransporterFacade & _facade):
   theStop(0),
   theFacade(_facade)
 {
+  DBUG_ENTER("ClusterMgr::ClusterMgr");
   ndbSetOwnVersion();
   clusterMgrThreadMutex = NdbMutex_Create();
   noOfAliveNodes= 0;
   noOfConnectedNodes= 0;
   theClusterMgrThread= 0;
+  DBUG_VOID_RETURN;
 }
 
-ClusterMgr::~ClusterMgr(){
+ClusterMgr::~ClusterMgr()
+{
+  DBUG_ENTER("ClusterMgr::~ClusterMgr");
   doStop();  
   NdbMutex_Destroy(clusterMgrThreadMutex);
+  DBUG_VOID_RETURN;
 }
 
 void
@@ -152,7 +157,6 @@ ClusterMgr::doStop( ){
   if (theClusterMgrThread) {
     NdbThread_WaitFor(theClusterMgrThread, &status);  
     NdbThread_Destroy(&theClusterMgrThread);
-    theClusterMgrThread= 0;
   }
   NdbMutex_Unlock(clusterMgrThreadMutex);
   DBUG_VOID_RETURN;
@@ -210,7 +214,7 @@ ClusterMgr::threadMain( ){
 	 * It is now time to send a new Heartbeat
 	 */
 	if (theNode.hbCounter >= theNode.hbFrequency) {
-	  theNode.hbSent++;
+	  theNode.m_info.m_heartbeat_cnt++;
 	  theNode.hbCounter = 0;
 	}
 
@@ -227,7 +231,7 @@ ClusterMgr::threadMain( ){
 	theFacade.sendSignalUnCond(&signal, nodeId);
       }//if
       
-      if (theNode.hbSent == 4 && theNode.hbFrequency > 0){
+      if (theNode.m_info.m_heartbeat_cnt == 4 && theNode.hbFrequency > 0){
 	reportNodeFailed(i);
       }//if
     }
@@ -261,6 +265,7 @@ ClusterMgr::Node::Node()
   : m_state(NodeState::SL_NOTHING) { 
   compatible = nfCompleteRep = true;
   connected = defined = m_alive = false; 
+  m_state.m_connected_nodes.clear();
 }
 
 /******************************************************************************
@@ -332,7 +337,7 @@ ClusterMgr::execAPI_REGCONF(const Uint32 * theData){
       node.compatible = ndbCompatible_api_ndb(NDB_VERSION,
 					      node.m_info.m_version);
   }
-  
+
   node.m_state = apiRegConf->nodeState;
   if (node.compatible && (node.m_state.startLevel == NodeState::SL_STARTED  ||
 			  node.m_state.startLevel == NodeState::SL_SINGLEUSER)){
@@ -340,7 +345,7 @@ ClusterMgr::execAPI_REGCONF(const Uint32 * theData){
   } else {
     set_node_alive(node, false);
   }//if
-  node.hbSent = 0;
+  node.m_info.m_heartbeat_cnt = 0;
   node.hbCounter = 0;
   if (node.m_info.m_type != NodeInfo::REP) {
     node.hbFrequency = (apiRegConf->apiHeartbeatFrequency * 10) - 50;
@@ -409,9 +414,15 @@ ClusterMgr::reportConnected(NodeId nodeId){
 
   Node & theNode = theNodes[nodeId];
   theNode.connected = true;
-  theNode.hbSent = 0;
+  theNode.m_info.m_heartbeat_cnt = 0;
   theNode.hbCounter = 0;
-  
+
+  /**
+   * make sure the node itself is marked connected even
+   * if first API_REGCONF has not arrived
+   */
+  theNode.m_state.m_connected_nodes.set(nodeId);
+
   if (theNode.m_info.m_type != NodeInfo::REP) {
     theNode.hbFrequency = 0;
   }
@@ -429,7 +440,9 @@ ClusterMgr::reportDisconnected(NodeId nodeId){
 
   noOfConnectedNodes--;
   theNodes[nodeId].connected = false;
-  theNodes[nodeId].m_info.m_connectCount ++;
+
+  theNodes[nodeId].m_state.m_connected_nodes.clear();
+
   reportNodeFailed(nodeId);
 }
 
@@ -439,19 +452,23 @@ ClusterMgr::reportNodeFailed(NodeId nodeId){
   Node & theNode = theNodes[nodeId];
  
   set_node_alive(theNode, false);
-  if(theNode.connected)
-    theFacade.doDisconnect(nodeId);
+  theNode.m_info.m_connectCount ++;
   
+  if(theNode.connected)
+  {
+    theFacade.doDisconnect(nodeId);
+  }
   const bool report = (theNode.m_state.startLevel != NodeState::SL_NOTHING);  
   theNode.m_state.startLevel = NodeState::SL_NOTHING;
   
-  if(report){
+  if(report)
+  {
     theFacade.ReportNodeDead(nodeId);
-  }  
-
-  theNode.nfCompleteRep = false;
+  }
   
-  if(noOfAliveNodes == 0){
+  theNode.nfCompleteRep = false;
+  if(noOfAliveNodes == 0)
+  {
     NFCompleteRep rep;
     for(Uint32 i = 1; i<MAX_NODES; i++){
       if(theNodes[i].defined && theNodes[i].nfCompleteRep == false){
@@ -468,6 +485,8 @@ ClusterMgr::reportNodeFailed(NodeId nodeId){
 ArbitMgr::ArbitMgr(TransporterFacade & _fac)
   : theFacade(_fac)
 {
+  DBUG_ENTER("ArbitMgr::ArbitMgr");
+
   theThreadMutex = NdbMutex_Create();
   theInputCond = NdbCondition_Create();
   theInputMutex = NdbMutex_Create();
@@ -485,13 +504,17 @@ ArbitMgr::ArbitMgr(TransporterFacade & _fac)
   memset(&theChooseReq1, 0, sizeof(theChooseReq1));
   memset(&theChooseReq2, 0, sizeof(theChooseReq2));
   memset(&theStopOrd, 0, sizeof(theStopOrd));
+
+  DBUG_VOID_RETURN;
 }
 
 ArbitMgr::~ArbitMgr()
 {
+  DBUG_ENTER("ArbitMgr::~ArbitMgr");
   NdbMutex_Destroy(theThreadMutex);
   NdbCondition_Destroy(theInputCond);
   NdbMutex_Destroy(theInputMutex);
+  DBUG_VOID_RETURN;
 }
 
 // Start arbitrator thread.  This is kernel request.
@@ -508,7 +531,7 @@ ArbitMgr::doStart(const Uint32* theData)
     sendSignalToThread(aSignal);
     void* value;
     NdbThread_WaitFor(theThread, &value);
-    theThread = NULL;
+    NdbThread_Destroy(&theThread);
     theState = StateInit;
     theInputFull = false;
   }
@@ -547,7 +570,7 @@ ArbitMgr::doStop(const Uint32* theData)
     sendSignalToThread(aSignal);
     void* value;
     NdbThread_WaitFor(theThread, &value);
-    theThread = NULL;
+    NdbThread_Destroy(&theThread);
     theState = StateInit;
   }
   NdbMutex_Unlock(theThreadMutex);
