@@ -30,7 +30,6 @@
 #define EILSEQ ENOENT
 #endif
 
-extern MY_UNICASE_INFO *uni_plane[256];
 
 static uchar ctype_ucs2[] = {
     0,
@@ -95,7 +94,7 @@ static int my_ucs2_uni(CHARSET_INFO *cs __attribute__((unused)),
 		       my_wc_t * pwc, const uchar *s, const uchar *e)
 {
   if (s+2 > e) /* Need 2 characters */
-    return MY_CS_TOOFEW(0);
+    return MY_CS_TOOSMALL2;
   
   *pwc= ((unsigned char)s[0]) * 256  + ((unsigned char)s[1]);
   return 2;
@@ -105,7 +104,7 @@ static int my_uni_ucs2(CHARSET_INFO *cs __attribute__((unused)) ,
 		       my_wc_t wc, uchar *r, uchar *e)
 {
   if ( r+2 > e ) 
-    return MY_CS_TOOSMALL;
+    return MY_CS_TOOSMALL2;
   
   r[0]= (uchar) (wc >> 8);
   r[1]= (uchar) (wc & 0xFF);
@@ -113,20 +112,26 @@ static int my_uni_ucs2(CHARSET_INFO *cs __attribute__((unused)) ,
 }
 
 
-static void my_caseup_ucs2(CHARSET_INFO *cs, char *s, uint slen)
+static uint my_caseup_ucs2(CHARSET_INFO *cs, char *src, uint srclen,
+                           char *dst __attribute__((unused)),
+                           uint dstlen __attribute__((unused)))
 {
   my_wc_t wc;
   int res;
-  char *e=s+slen;
-
-  while ((s < e) && (res=my_ucs2_uni(cs,&wc, (uchar *)s, (uchar*)e))>0 )
+  char *srcend= src + srclen;
+  MY_UNICASE_INFO **uni_plane= cs->caseinfo;
+  DBUG_ASSERT(src == dst && srclen == dstlen);
+  
+  while ((src < srcend) &&
+         (res= my_ucs2_uni(cs, &wc, (uchar *)src, (uchar*) srcend)) > 0)
   {
-    int plane = (wc>>8) & 0xFF;
-    wc = uni_plane[plane] ? uni_plane[plane][wc & 0xFF].toupper : wc;
-    if (res != my_uni_ucs2(cs,wc,(uchar*)s,(uchar*)e))
+    int plane= (wc>>8) & 0xFF;
+    wc= uni_plane[plane] ? uni_plane[plane][wc & 0xFF].toupper : wc;
+    if (res != my_uni_ucs2(cs, wc, (uchar*) src, (uchar*) srcend))
       break;
-    s+=res;
+    src+= res;
   }
+  return srclen;
 }
 
 
@@ -136,6 +141,10 @@ static void my_hash_sort_ucs2(CHARSET_INFO *cs, const uchar *s, uint slen,
   my_wc_t wc;
   int res;
   const uchar *e=s+slen;
+  MY_UNICASE_INFO **uni_plane= cs->caseinfo;
+
+  while (e > s+1 && e[-1] == ' ' && e[-2] == '\0')
+    e-= 2;
 
   while ((s < e) && (res=my_ucs2_uni(cs,&wc, (uchar *)s, (uchar*)e)) >0)
   {
@@ -157,22 +166,26 @@ static void my_caseup_str_ucs2(CHARSET_INFO * cs  __attribute__((unused)),
 
 
 
-static void my_casedn_ucs2(CHARSET_INFO *cs, char *s, uint slen)
+static uint my_casedn_ucs2(CHARSET_INFO *cs, char *src, uint srclen,
+                           char *dst __attribute__((unused)),
+                           uint dstlen __attribute__((unused)))
 {
   my_wc_t wc;
   int res;
-  char *e=s+slen;
+  char *srcend= src + srclen;
+  MY_UNICASE_INFO **uni_plane= cs->caseinfo;
+  DBUG_ASSERT(src == dst && srclen == dstlen);
 
-  while ((s < e) && (res=my_ucs2_uni(cs, &wc, (uchar*)s, (uchar*)e))>0)
+  while ((src < srcend) &&
+         (res= my_ucs2_uni(cs, &wc, (uchar*) src, (uchar*) srcend)) > 0)
   {
-    int plane = (wc>>8) & 0xFF;
-    wc = uni_plane[plane] ? uni_plane[plane][wc & 0xFF].tolower : wc;
-    if (res != my_uni_ucs2(cs, wc, (uchar*)s, (uchar*)e))
-    {
+    int plane= (wc>>8) & 0xFF;
+    wc= uni_plane[plane] ? uni_plane[plane][wc & 0xFF].tolower : wc;
+    if (res != my_uni_ucs2(cs, wc, (uchar*) src, (uchar*) srcend))
       break;
-    }
-    s+=res;
+    src+= res;
   }
+  return srclen;
 }
 
 static void my_casedn_str_ucs2(CHARSET_INFO *cs __attribute__((unused)), 
@@ -190,6 +203,7 @@ static int my_strnncoll_ucs2(CHARSET_INFO *cs,
   my_wc_t s_wc,t_wc;
   const uchar *se=s+slen;
   const uchar *te=t+tlen;
+  MY_UNICASE_INFO **uni_plane= cs->caseinfo;
 
   while ( s < se && t < te )
   {
@@ -215,7 +229,7 @@ static int my_strnncoll_ucs2(CHARSET_INFO *cs,
     s+=s_res;
     t+=t_res;
   }
-  return t_is_prefix ? (int) (t - te) : (int) ((se - s) - (te - t));
+  return (int) (t_is_prefix ? t-te : ((se-s) - (te-t)));
 }
 
 /*
@@ -247,14 +261,17 @@ static int my_strnncoll_ucs2(CHARSET_INFO *cs,
 
 static int my_strnncollsp_ucs2(CHARSET_INFO *cs __attribute__((unused)),
                                const uchar *s, uint slen,
-                               const uchar *t, uint tlen)
+                               const uchar *t, uint tlen,
+                               my_bool diff_if_only_endspace_difference
+			       __attribute__((unused)))
 {
   const uchar *se, *te;
   uint minlen;
+  MY_UNICASE_INFO **uni_plane= cs->caseinfo;
 
   /* extra safety to make sure the lengths are even numbers */
-  slen= (slen >> 1) << 1;
-  tlen= (tlen >> 1) << 1;
+  slen&= ~1;
+  tlen&= ~1;
 
   se= s + slen;
   te= t + tlen;
@@ -300,6 +317,7 @@ static int my_strncasecmp_ucs2(CHARSET_INFO *cs,
   my_wc_t s_wc,t_wc;
   const char *se=s+len;
   const char *te=t+len;
+  MY_UNICASE_INFO **uni_plane= cs->caseinfo;
   
   while ( s < se && t < te )
   {
@@ -332,8 +350,8 @@ static int my_strncasecmp_ucs2(CHARSET_INFO *cs,
 
 static int my_strcasecmp_ucs2(CHARSET_INFO *cs, const char *s, const char *t)
 {
-  uint s_len=strlen(s);
-  uint t_len=strlen(t);
+  uint s_len= (uint) strlen(s);
+  uint t_len= (uint) strlen(t);
   uint len = (s_len > t_len) ? s_len : t_len;
   return  my_strncasecmp_ucs2(cs, s, t, len);
 }
@@ -347,6 +365,7 @@ static int my_strnxfrm_ucs2(CHARSET_INFO *cs,
   int plane;
   uchar *de = dst + dstlen;
   const uchar *se = src + srclen;
+  MY_UNICASE_INFO **uni_plane= cs->caseinfo;
 
   while( src < se && dst < de )
   {
@@ -367,7 +386,7 @@ static int my_strnxfrm_ucs2(CHARSET_INFO *cs,
     dst+=res;
   }
   if (dst < de)
-    cs->cset->fill(cs, dst, de - dst, ' ');
+    cs->cset->fill(cs, (char*) dst, (uint) (de - dst), ' ');
   return dstlen;
 }
 
@@ -921,15 +940,16 @@ bs:
   return (negative ? -((longlong) res) : (longlong) res);
 }
 
-double      my_strntod_ucs2(CHARSET_INFO *cs __attribute__((unused)),
-			   char *nptr, uint length, 
-			   char **endptr, int *err)
+
+double my_strntod_ucs2(CHARSET_INFO *cs __attribute__((unused)),
+                       char *nptr, uint length, 
+                       char **endptr, int *err)
 {
   char     buf[256];
   double   res;
   register char *b=buf;
   register const uchar *s= (const uchar*) nptr;
-  register const uchar *end;
+  const uchar *end;
   my_wc_t  wc;
   int      cnv;
 
@@ -1043,7 +1063,7 @@ int my_ll10tostr_ucs2(CHARSET_INFO *cs __attribute__((unused)),
   while (long_val != 0)
   {
     long quo= long_val/10;
-    *--p = '0' + (long_val - quo*10);
+    *--p = (char) ('0' + (long_val - quo*10));
     long_val= quo;
   }
   
@@ -1304,6 +1324,7 @@ int my_wildcmp_ucs2_ci(CHARSET_INFO *cs,
 		    const char *wildstr,const char *wildend,
 		    int escape, int w_one, int w_many)
 {
+  MY_UNICASE_INFO **uni_plane= cs->caseinfo;
   return my_wildcmp_unicode(cs,str,str_end,wildstr,wildend,
                             escape,w_one,w_many,uni_plane); 
 }
@@ -1349,22 +1370,61 @@ int my_strnncoll_ucs2_bin(CHARSET_INFO *cs,
     s+=s_res;
     t+=t_res;
   }
-  return t_is_prefix ? (int) (t - te) : (int) ((se-s) - (te-t));
+  return (int) (t_is_prefix ? t-te : ((se-s) - (te-t)));
 }
 
-static int my_strnncollsp_ucs2_bin(CHARSET_INFO *cs, 
+static int my_strnncollsp_ucs2_bin(CHARSET_INFO *cs __attribute__((unused)), 
                                    const uchar *s, uint slen, 
-                                   const uchar *t, uint tlen)
+                                   const uchar *t, uint tlen,
+                                   my_bool diff_if_only_endspace_difference
+                                   __attribute__((unused)))
 {
-  return my_strnncoll_ucs2_bin(cs,s,slen,t,tlen,0);
+  const uchar *se, *te;
+  uint minlen;
+
+  /* extra safety to make sure the lengths are even numbers */
+  slen= (slen >> 1) << 1;
+  tlen= (tlen >> 1) << 1;
+
+  se= s + slen;
+  te= t + tlen;
+
+  for (minlen= min(slen, tlen); minlen; minlen-= 2)
+  {
+    int s_wc= s[0] * 256 + s[1];
+    int t_wc= t[0] * 256 + t[1];
+    if ( s_wc != t_wc )
+      return  s_wc > t_wc ? 1 : -1;
+
+    s+= 2;
+    t+= 2;
+  }
+
+  if (slen != tlen)
+  {
+    int swap= 1;
+    if (slen < tlen)
+    {
+      s= t;
+      se= te;
+      swap= -1;
+    }
+
+    for ( ; s < se ; s+= 2)
+    {
+      if (s[0] || s[1] != ' ')
+        return (s[0] == 0 && s[1] < ' ') ? -swap : swap;
+    }
+  }
+  return 0;
 }
 
 
 static
 int my_strcasecmp_ucs2_bin(CHARSET_INFO *cs, const char *s, const char *t)
 {
-  uint s_len=strlen(s);
-  uint t_len=strlen(t);
+  uint s_len= (uint) strlen(s);
+  uint t_len= (uint) strlen(t);
   uint len = (s_len > t_len) ? s_len : t_len;
   return  my_strncasecmp_ucs2(cs, s, t, len);
 }
@@ -1378,7 +1438,7 @@ int my_strnxfrm_ucs2_bin(CHARSET_INFO *cs __attribute__((unused)),
   if (dst != src)
     memcpy(dst,src,srclen= min(dstlen,srclen));
   if (dstlen > srclen)
-    cs->cset->fill(cs, dst + srclen, dstlen - srclen, ' ');
+    cs->cset->fill(cs, (char*) dst + srclen, dstlen - srclen, ' ');
   return dstlen;
 }
 
@@ -1448,8 +1508,14 @@ my_bool my_like_range_ucs2(CHARSET_INFO *cs,
     }
     if (ptr[0] == '\0' && ptr[1] == w_many)	/* '%' in SQL */
     {
-      *min_length= (uint) (min_str - min_org);
-      *max_length=res_length;
+      /*
+        Calculate length of keys:
+        'a\0\0... is the smallest possible string when we have space expand
+        a\ff\ff... is the biggest possible string
+      */
+      *min_length= ((cs->state & MY_CS_BINSORT) ? (uint) (min_str - min_org) :
+                    res_length);
+      *max_length= res_length;
       do {
         *min_str++ = 0;
 	*min_str++ = 0;
@@ -1461,7 +1527,6 @@ my_bool my_like_range_ucs2(CHARSET_INFO *cs,
     *min_str++= *max_str++ = ptr[0];
     *min_str++= *max_str++ = ptr[1];
   }
-  *min_length= *max_length = (uint) (min_str - min_org);
 
   /* Temporary fix for handling w_one at end of string (key compression) */
   {
@@ -1473,15 +1538,17 @@ my_bool my_like_range_ucs2(CHARSET_INFO *cs,
     }
   }
   
+  *min_length= *max_length = (uint) (min_str - min_org);
   while (min_str + 1 < min_end)
   {
     *min_str++ = *max_str++ = '\0';
-    *min_str++ = *max_str++ = ' ';	/* Because if key compression */
+    *min_str++ = *max_str++ = ' ';      /* Because if key compression */
   }
   return 0;
 }
 
 
+
 ulong my_scan_ucs2(CHARSET_INFO *cs __attribute__((unused)),
                    const char *str, const char *end, int sequence_type)
 {
@@ -1510,11 +1577,13 @@ static MY_COLLATION_HANDLER my_collation_ucs2_general_ci_handler =
     my_strnncoll_ucs2,
     my_strnncollsp_ucs2,
     my_strnxfrm_ucs2,
+    my_strnxfrmlen_simple,
     my_like_range_ucs2,
     my_wildcmp_ucs2_ci,
     my_strcasecmp_ucs2,
     my_instr_mb,
-    my_hash_sort_ucs2
+    my_hash_sort_ucs2,
+    my_propagate_simple
 };
 
 
@@ -1524,11 +1593,13 @@ static MY_COLLATION_HANDLER my_collation_ucs2_bin_handler =
     my_strnncoll_ucs2_bin,
     my_strnncollsp_ucs2_bin,
     my_strnxfrm_ucs2_bin,
+    my_strnxfrmlen_simple,
     my_like_range_simple,
     my_wildcmp_ucs2_bin,
     my_strcasecmp_ucs2_bin,
     my_instr_mb,
-    my_hash_sort_ucs2_bin
+    my_hash_sort_ucs2_bin,
+    my_propagate_simple
 };
 
 
@@ -1578,13 +1649,17 @@ CHARSET_INFO my_charset_ucs2_general_ci=
     NULL,		/* sort_order_big*/
     NULL,		/* tab_to_uni   */
     NULL,		/* tab_from_uni */
+    my_unicase_default, /* caseinfo     */
     NULL,		/* state_map    */
     NULL,		/* ident_map    */
     1,			/* strxfrm_multiply */
+    1,                  /* caseup_multiply  */
+    1,                  /* casedn_multiply  */
     2,			/* mbminlen     */
     2,			/* mbmaxlen     */
     0,			/* min_sort_char */
     0xFFFF,		/* max_sort_char */
+    ' ',                /* pad char      */
     0,                  /* escape_with_backslash_is_dangerous */
     &my_charset_ucs2_handler,
     &my_collation_ucs2_general_ci_handler
@@ -1606,13 +1681,17 @@ CHARSET_INFO my_charset_ucs2_bin=
     NULL,		/* sort_order_big*/
     NULL,		/* tab_to_uni   */
     NULL,		/* tab_from_uni */
+    my_unicase_default, /* caseinfo     */
     NULL,		/* state_map    */
     NULL,		/* ident_map    */
     1,			/* strxfrm_multiply */
+    1,                  /* caseup_multiply  */
+    1,                  /* casedn_multiply  */
     2,			/* mbminlen     */
     2,			/* mbmaxlen     */
     0,			/* min_sort_char */
     0xFFFF,		/* max_sort_char */
+    ' ',                /* pad char      */
     0,                  /* escape_with_backslash_is_dangerous */
     &my_charset_ucs2_handler,
     &my_collation_ucs2_bin_handler
